//Sea.java - Part of Bill's Java Minesweeper
import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.border.*;
import java.net.*;

public class Sea {
   JPanel seaGrid;
   SeaButton[][] seaArray;
   JTextArea debugText;
   ArrayList buttonList,mineList;
   AdjacentCells adjCells;
   MineCnt mineCount;
   Smiley smiley;
   MyMines myMines;
   URL codeBase;
   //Colored numbers for adjacent mine count.
   static Color[] numberColors = new Color[] {
      Color.black,   //Dummy for 0.
      Color.blue,    //1
      new Color(60,140,40), //2 (forest green)
      Color.red,     //3
      Color.black,   //4
      new Color(220,100,0), //5 (bolder orange)
      Color.magenta, //6
      Color.cyan,    //7
      Color.yellow   //8
   };

   public Sea(MyMines myMines, JTextArea debugText, MineCnt mineCount, Smiley smiley) {
      this.myMines = myMines;
      this.debugText = debugText;
      this.mineCount = mineCount;
      this.smiley = smiley;
      codeBase = myMines.getCodeBase();
      adjCells = new AdjacentCells(this);
      seaGrid = new JPanel( new GridLayout(Game.rows,Game.cols,0,0) );
      seaArray = new SeaButton[Game.rows][Game.cols];
      buttonList = new ArrayList();
      mineList = new ArrayList();
      for(int outerLoop=0; outerLoop<Game.rows; outerLoop++) {
         for(int innerLoop=0; innerLoop<Game.cols; innerLoop++) {
            seaArray[outerLoop][innerLoop] = new SeaButton(outerLoop,innerLoop);
            buttonList.add(seaArray[outerLoop][innerLoop]);
            seaGrid.add(seaArray[outerLoop][innerLoop]);
         }
      }
      buttonList.trimToSize();
      setMines(Game.mines);
   }

   public JPanel getJPanel() {return seaGrid;}

   class SeaButton extends JButton {
      int row, col;
      String buttonLabel;
      private boolean
         revealed = false,
         marked = false,
         questionMarked = false,
         mined = false;
      int adjacentMines = 0;
      Border bevelBorder,grayLineBorder;
      ImageIcon flag,mine1;

      SeaButton(int row, int col) {
         this.row = row;
         this.col = col;
         buttonLabel = new String("r,c = (" + row + ", " + col + ")  ");
         setMargin(new Insets(0,0,0,0));
         setPreferredSize(new Dimension(17,17));
         addMouseListener(new ClickListener(debugText, this));
         bevelBorder = BorderFactory.createBevelBorder(BevelBorder.RAISED);
         setBorder(bevelBorder);
         grayLineBorder = BorderFactory.createLineBorder(Color.gray);
         flag = new ImageIcon(myMines.getImage(codeBase, "flag.gif"));
         mine1 = new ImageIcon(myMines.getImage(codeBase, "mine1.gif"));
      }

      boolean isMined() { return mined; }
      boolean isRevealed() { return revealed; }
      boolean isMarked() { return marked; }
      boolean isQuestionMarked() { return questionMarked; }

      void placeMine() {
         mined = true;
         adjacentMines = -1;
      }

      void setAdjacentMines(int mines) {
         adjacentMines = mines;
      }

      void reveal() {
         //debugText.append("Game.started: " + Game.started + "\n");
         if(!revealed && !Game.over) {
            if(marked) {
               //Call worried smiley here.  Do not reveal.
               smiley.setWorry();
               debugText.append("Worry: Can't reveal!\n");
            } else {
               revealed = true;
               setBorder(grayLineBorder);
               if(mined) {
                  loseGame();
               }else if(adjacentMines == 0) {
                  //Note that the process of expanding zero regions is recursive.
                  adjCells.findZeros(row, col);
               }else if(adjacentMines > 0) {
                  setForeground(numberColors[adjacentMines]);
                  setText(Integer.toString(adjacentMines));
               }
               int index = buttonList.lastIndexOf(this);
               if(index >= 0) {
                  buttonList.remove(index);
                  debugText.append("Button list size: " + buttonList.size() + "\n");
                  checkForWin();
               }
            }
            /* If player bothClicked mistakenly, the background color was changed to
               highlight the error.  Return it to normal color now. */
            if(getBackground() == Color.gray) setBackground(Color.lightGray);
         }
      }//End reveal.

      void exposeMine() {
         if(revealed) {
            // This is the one that lost the game.
            setBackground(Color.red);
         }
         if(!marked) {
            revealed = true;
            setBorder(grayLineBorder);
            setIcon(mine1);
         }
      }//End exposeMine.

      void revealAdjacent() {
         //For starters, just check whether there are unmarked mines.
         debugText.append("Unmarked Mines: " +
            adjCells.isUnmarkedMines(row, col) + "\n");
      }

      void mark() {
         if(!revealed && !Game.over) {
            if(marked) {
               marked = false;
               mineCount.increment();  //Removing a mark increases the count of unmarked mines.
               if(Game.questionMarks) {
                  questionMarked = true;
                  setIcon(null);
                  setText("?");
               } else {
                  setIcon(null);
               }
            } else if(questionMarked) {
               questionMarked = false;
               setText(null);
            } else {
               marked = true;
               setIcon(flag);
               mineCount.decrement();
               checkForWin();
            }
            /* If player bothClicked mistakenly, the background color was changed to
               highlight the error.  Return it to normal color now. */
            if(getBackground() == Color.gray) setBackground(Color.lightGray);
         }//End !revealed.
      }//End mark.

   }//End SeaButton.

   void setMines(int numMines) {
      Random rand = new Random();
      int index;
      SeaButton aButton;
      for(int loop=0; loop<numMines; loop++) {
         index = rand.nextInt(buttonList.size());
         aButton = (SeaButton)buttonList.remove(index);
         mineList.add(aButton);
         aButton.placeMine();
      }
      //After placing mines, buttonList contains all non-mined SeaButtons.
      //mineList contains all mined SeaButtons.
      //For each, count mines in adjacent cells.
      ListIterator li = buttonList.listIterator();
      int mineCount;
      //AdjacentCells adjCells = new AdjacentCells(Sea.this);
      while(li.hasNext()) {
         aButton = (SeaButton)li.next();
         mineCount = adjCells.countMines(aButton.row, aButton.col);
         aButton.setAdjacentMines(mineCount);
      }
   }//End setMines.

   void loseGame() {
      Game.over = true;
      smiley.setFrown();
      debugText.append("Frown: You lose!!\n");

      //Show all unmarked mines.
      ListIterator li = mineList.listIterator();
      SeaButton aButton;
      while(li.hasNext()) {
         aButton = (SeaButton)li.next();
         aButton.exposeMine();
      }
   }//End loseGame.

   void checkForWin() {
      if(buttonList.size() == 0) {
         if(mineCount.read() == 0) {
            debugText.append("Cool: You won!!\n");
            Game.over = true;
            smiley.setShades();
         }
      }
   }//End checkForWin.

}//End Sea.

class AdjacentCells {
   Sea mySea;
   AdjacentCells(Sea mySea) { this.mySea = mySea; }

   ArrayList getList(int row, int col) {
      ArrayList adjCells = new ArrayList(8);
      if(isInBounds(row-1, col  )) adjCells.add(mySea.seaArray[row-1][col  ]); //N
      if(isInBounds(row-1, col+1)) adjCells.add(mySea.seaArray[row-1][col+1]); //NE
      if(isInBounds(row  , col+1)) adjCells.add(mySea.seaArray[row  ][col+1]); //E
      if(isInBounds(row+1, col+1)) adjCells.add(mySea.seaArray[row+1][col+1]); //SE
      if(isInBounds(row+1, col  )) adjCells.add(mySea.seaArray[row+1][col  ]); //S
      if(isInBounds(row+1, col-1)) adjCells.add(mySea.seaArray[row+1][col-1]); //SW
      if(isInBounds(row  , col-1)) adjCells.add(mySea.seaArray[row  ][col-1]); //W
      if(isInBounds(row-1, col-1)) adjCells.add(mySea.seaArray[row-1][col-1]); //NW
      adjCells.trimToSize();
      return adjCells;
   }

   int countMines(int row, int col) {
      ArrayList adjCells = getList(row, col);
      ListIterator li = adjCells.listIterator();
      Sea.SeaButton sb;
      int count = 0;
      while(li.hasNext()) {
         sb = (Sea.SeaButton)li.next();
         if(sb.isMined()) count++;
      }
      return count;
   }

   boolean isUnmarkedMines(int row, int col) {
      /* When player double clicks a numbered cell, if all adjacent mines are marked,
         remaining adjacent cells are revealed. */
      boolean unmarkedMines = false;
      ArrayList adjCells = getList(row, col);
      ListIterator li = adjCells.listIterator();
      Sea.SeaButton sb;
      while(li.hasNext()) {
         sb = (Sea.SeaButton)li.next();
         if( sb.isMined() && !sb.isMarked() ) unmarkedMines = true;
         if( sb.isRevealed() || sb.isMarked() ) li.remove();
      }

      /* Now we know whether there are unmarked mines, and we have a list of
         unrevealed adjacent buttons.  If no unmarked mines, reveal these buttons.
         If unmarked mines, highlight the unrevealed buttons. */
      while(li.hasPrevious()) {
         sb = (Sea.SeaButton)li.previous();
         if(unmarkedMines) sb.setBackground(Color.gray);
         else sb.reveal();
      }

      return unmarkedMines;
   }

   void findZeros(int row, int col) {
      ArrayList adjCells = getList(row, col);
      ListIterator li = adjCells.listIterator();
      Sea.SeaButton sb;
      while(li.hasNext()) {
         sb = (Sea.SeaButton)li.next();
         if(sb.adjacentMines >= 0)
            if(!sb.isRevealed())
               sb.reveal();      //Note recursion here.
      }
   }

   boolean isInBounds(int row, int col) {
      if(row >= 0  &&  row < Game.rows  &&
         col >= 0  &&  col < Game.cols      ) { return true; }
      else { return false; }
   }
}
